"""3DSliceImages: Efficiently generate and store 2D slices from a 3D SimpleITK image."""

# %%
from __future__ import annotations

import base64
import time
from dataclasses import dataclass, field
from io import BytesIO
from typing import TYPE_CHECKING, Iterator

import ipywidgets as widgets  # type: ignore
import numpy as np
from IPython.display import display
from PIL import Image
from tqdm.notebook import tqdm

from imgtools.loggers import logger
from imgtools.utils import image_to_array

if TYPE_CHECKING:
    from pathlib import Path

    import SimpleITK as sitk

__all__: list[str] = [
    "ImageSlices",
    "SliceImage3D",
    "view_multiple_SliceImage3DObjects",
]


@dataclass
class ImageSlices:
    image_list: list[Image.Image] = field(default_factory=list)

    @classmethod
    def from_dict(
        cls, slices: dict[tuple[int, int], Image.Image]
    ) -> "ImageSlices":
        image_list = [slices[key] for key in sorted(slices.keys())]
        return cls(image_list)

    def export_gif(self, output_path: Path) -> Path:
        self.image_list[0].save(
            output_path,
            save_all=True,
            append_images=self.image_list[1:],
            duration=100,
            loop=0,
        )
        return output_path

    def export_pngs(self, output_dir: Path) -> list[Path]:
        output_dir.mkdir(parents=True, exist_ok=True)
        paths = []
        for i, image in enumerate(self.image_list):
            file_path = output_dir / f"slice_{i}.png"
            image.save(file_path)
            paths.append(file_path)
        return paths

    def __getitem__(self, index: int) -> Image.Image:
        return self.image_list[index]

    def __len__(self) -> int:
        return len(self.image_list)

    def __iter__(self) -> Iterator[Image.Image]:
        return iter(self.image_list)


@dataclass
class SliceImage3D:
    """Generates 2D slices from a 3D SimpleITK image and optionally overlays a mask."""

    image: sitk.Image
    mask: sitk.Image | None = None
    alpha: float = 0.2
    disable_progress: bool = False

    slices: ImageSlices = field(init=False)
    image_array: np.ndarray = field(init=False)
    mask_array: np.ndarray | None = field(init=False)

    def __post_init__(self) -> None:
        logger.info("Original image shape: " + str(self.image.GetSize()))

        self.image_array, *_image = image_to_array(self.image)

        if self.mask:
            self.mask_array, *_mask = image_to_array(self.mask)
        else:
            self.mask_array = None

        logger.info(f"Image shape: {self.image_array.shape}")
        if self.mask_array is not None:
            logger.info(f"Mask shape: {self.mask_array.shape}")

    @property
    def vmax(self) -> int:
        return max(
            self.image_array.max(),
            self.mask_array.max() if self.mask_array is not None else 0,
        )

    @property
    def vmin(self) -> int:
        return min(
            self.image_array.min(),
            self.mask_array.min() if self.mask_array is not None else 0,
        )

    def image_slices(self, dim: int, every_n: int = 5) -> list[np.ndarray]:
        return [
            np.take(self.image_array, i, axis=dim)
            for i in range(0, self.image_array.shape[dim], every_n)
        ]

    def mask_slices(
        self, dim: int, every_n: int = 5
    ) -> list[np.ndarray] | None:
        if self.mask_array is not None:
            return [
                np.take(self.mask_array, i, axis=dim)
                for i in range(0, self.mask_array.shape[dim], every_n)
            ]
        return None

    def generate_dim_slices(
        self,
        dim: int,
        every_n: int = 5,
    ) -> SliceImage3D:
        """Generate 2D slices along a given dimension, optionally overlaying the mask.

        Parameters
        ----------
        dim : int
            The dimension along which to generate slices (0, 1, or 2).
        every_n : int, optional
            Generate a slice every `every_n` pixels, by default 5.
        """
        slices = self.image_slices(dim, every_n)
        mask_slices: list[np.ndarray] | None = self.mask_slices(dim, every_n)

        if mask_slices is None:
            mask_slices = [
                np.zeros_like(slices[0]) for _ in range(len(slices))
            ]
        logger.warning("Generating slices...")
        tasks = [
            (
                i,
                img,
                mask,
                self.alpha,
                self.vmax,
                self.vmin,
            )
            for i, (img, mask) in enumerate(
                zip(slices, mask_slices, strict=False)
            )
        ]
        start = time.time()
        results = [
            SliceImage3D._generate_slice(*task)
            for task in tqdm(tasks, disable=self.disable_progress)
        ]
        logger.info(f"Time taken: {time.time() - start:.2f}s")

        # Store slices in a dictionary
        images = [image for _, image in results]
        self.slices = ImageSlices(images)

        return self

    @staticmethod
    def _generate_slice(
        index: int,
        img_slice: np.ndarray,
        mask_slice: np.ndarray | None,
        alpha: float,
        vmax: int,
        vmin: int,
    ) -> tuple[int, Image.Image]:
        """Generate a single 2D slice, overlaying the mask in red if available,
        with metadata text in the bottom right corner.
        """

        # Normalize the image to 0-255
        img_slice = np.clip(
            (img_slice - vmin) / (vmax - vmin) * 255, 0, 255
        ).astype(np.uint8)

        # Convert image slice to PIL image (grayscale → RGB)
        image = Image.fromarray(img_slice, mode="L").convert("RGB")

        if mask_slice is not None:
            # mask should be all 1s and 0s
            red_mask = np.zeros_like(mask_slice, dtype=np.uint8)
            blue_mask = np.zeros_like(mask_slice, dtype=np.uint8)

            green_mask = np.clip(mask_slice * 255, 0, 255).astype(np.uint8)

            # Stack the channels to create an RGB image
            mask_img_stack = np.stack(
                [red_mask, green_mask, blue_mask], axis=-1
            )

            # Convert to PIL image
            mask_img = Image.fromarray(mask_img_stack, mode="RGB")

            # Create an alpha transparency mask where the mask_image is not black and multiply by alpha
            mask_alpha = Image.fromarray(
                np.clip(mask_slice * 255 * alpha, 0, 255).astype(np.uint8),
                mode="L",
            )

            # Blend the red mask with the original image
            image = Image.composite(mask_img, image, mask_alpha)

        return index, image

    def __repr__(self) -> str:
        slices_by_dimension = {
            dim: len(self.image_slices(dim)) for dim in range(3)
        }
        return (
            f"SliceImage3D(image_shape={self.image_array.shape}, "
            f"slices_by_dimension={slices_by_dimension})"
        )

    def view(self) -> None:
        def display_slice(index: int) -> None:
            img = self.slices[index]
            display(img)

        max_slider = len(self.slices) - 1
        widgets.interact(
            display_slice,
            index=widgets.IntSlider(
                min=0, max=max_slider, step=1, value=max_slider // 2
            ),
        )


def view_multiple_SliceImage3DObjects(*args: tuple[str, SliceImage3D]) -> None:  # noqa
    """This is a grand view to help when you want to view multiple

    pass in all of them, with generate_dim_slices already called
    ALL of them MUST have the same number of slices
    this will then use a SINGLE slider, to view and iterate over all of them

    """

    # calculate some metadata for all the slices
    metadata_dict = {
        name: f"Shape : x={slice_image.image_array.shape[2]}, y={slice_image.image_array.shape[1]}"
        for name, slice_image in args
    }

    def display_slices(index: int) -> None:
        # Create a vertical box layout for each slice with a label
        vboxes = []
        for name, slice_image in args:
            # Convert the Pillow image to Base64
            buffer = BytesIO()
            slice_image.slices[index].save(buffer, format="PNG")
            base64_image = base64.b64encode(buffer.getvalue()).decode()

            metadata_str = metadata_dict[name]
            # Generate HTML content with Base64-encoded image
            html_content = f"""
            <div style="text-align: center;">
                <h3>{name}</h3>
                <p>{metadata_str}</p>
                <img src="data:image/png;base64,{base64_image}" 
                    style="width: 500px; height: auto;">
            </div>
            """
            vbox = widgets.VBox([widgets.HTML(html_content)])
            vboxes.append(vbox)

        # Create a horizontal box layout for all vboxes
        hbox = widgets.HBox(vboxes)
        display(hbox)

    if not args:
        raise ValueError("At least one SliceImage3D object must be provided")

    num_slices = len(args[0][1].slices)
    if not all(
        len(slice_image.slices) == num_slices for _, slice_image in args
    ):
        errmsg = (
            "All SliceImage3D objects must have the same number of slices!!"
            f" (expected {num_slices}) but got {[len(s.slices) for _, s in args]}"
        )
        raise ValueError(errmsg)

    max_slider = num_slices - 1
    widgets.interact(
        display_slices,
        index=widgets.IntSlider(
            min=0, max=max_slider, step=1, value=max_slider // 2
        ),
    )


# %%
if __name__ == "__main__":  # pragma: no cover
    from rich import print, progress  # type: ignore # noqa
    from imgtools.coretypes.box import RegionBox, Size3D
    from imgtools.datasets.examples import data_images

    # from imgtools.ops.functional import resize

    logger.setLevel("DEBUG")  # type: ignore

    setting = 1
    dim = 0
    every = 5

    # Load example images
    ct_image = data_images()["duck"]
    seg_image = data_images()["mask"]

    # ct_image = resize(
    #     ct_image, size=np.array([512, 512, 0]), interpolation="linear"
    # )
    # seg_image = resize(
    #     seg_image, size=np.array([512, 512, 0]), interpolation="nearest"
    # )

    match setting:
        case 1:  # Crop to the bounding box of the mask
            bbox = RegionBox.from_mask_bbox(seg_image).expand_to_min_size(25)
            cropped_image = bbox.crop_image(ct_image)
            cropped_mask = bbox.crop_image(seg_image)
        case 2:  # Use the full image
            cropped_image = ct_image
            cropped_mask = seg_image

    new_size = 512
    if dim == 0:
        sz = Size3D(new_size, new_size, 0)
    elif dim == 1:
        sz = Size3D(new_size, 0, new_size)

    sz = Size3D(new_size, new_size, new_size)

    slices = SliceImage3D(
        image=cropped_image,
        mask=cropped_mask,
    )

    slices.generate_dim_slices(dim, every_n=every).view()


# %%



import datetime
import tempfile
from pathlib import Path

import pytest
from pydicom.dataset import Dataset, FileMetaDataset
from pydicom.uid import UID, ExplicitVRLittleEndian

from imgtools.dicom.sort.sorter_base import resolve_path


@pytest.fixture(scope='module')
def dicom_test_file():
    """Pytest fixture to create a DICOM file for testing."""
    ds = Dataset()
    # Patient and instance-related metadata
    ds.PatientName = 'Test^Firstname'
    ds.PatientID = '123456'

    # Dates and times
    datetime.datetime.now()
    ds.ContentDate = '20021114'
    ds.ContentTime = '111131'  # Updated time format

    # Other metadata
    ds.SpecificCharacterSet = 'ISO_IR 100'
    ds.ImageType = ['ORIGINAL', 'PRIMARY', 'AXIAL']
    ds.StudyDate = '20021114'
    ds.SeriesDate = '20021114'
    ds.AcquisitionDate = '20021114'
    ds.StudyTime = '105444'
    ds.SeriesTime = '110039'
    ds.AcquisitionTime = '110234.982284'
    ds.Modality = 'CT'
    ds.Manufacturer = 'GE MEDICAL SYSTEMS'
    ds.StudyDescription = 'CT ABD & PELVIS W/O &'
    ds.SeriesDescription = '2.5SOFT + 30%ASIR'
    ds.ManufacturerModelName = 'LightSpeed VCT'
    ds.SeriesInstanceUID = UID('1.2.840.113619.2.55.3.604688.12345678.1234567890')
    ds.StudyInstanceUID = UID('1.2.840.113619.2.55.3.604688.98765432.9876543210')
    # File meta information
    file_meta = FileMetaDataset()
    file_meta.MediaStorageSOPClassUID = UID('1.2.840.10008.5.1.4.1.1.2')
    file_meta.MediaStorageSOPInstanceUID = UID(
        '1.3.6.1.4.1.14519.5.2.1.1706.4016.292639580813240923432069920621'
    )
    file_meta.ImplementationClassUID = UID('1.2.3.4')
    file_meta.TransferSyntaxUID = ExplicitVRLittleEndian

    ds.file_meta = file_meta

    # Create a temporary file
    temp_file = Path(tempfile.NamedTemporaryFile(suffix='.dcm', delete=False).name)
    ds.save_as(temp_file, enforce_file_format=True)

    yield temp_file

    # Cleanup
    temp_file.unlink()


class TestResolvePath:


    def test_resolve_path_with_absolute_path_no_truncate(self, dicom_test_file) -> None:
        path = dicom_test_file
        keys = {'PatientID', 'StudyInstanceUID'}
        format_str = '/resolved/path/%(PatientID)s/%(StudyInstanceUID)s'
        result = resolve_path(path, keys, format_str, truncate=0)
        assert result[0].resolve() == path.resolve()
        assert result[1].resolve() == Path(
            format_str
            % {
                'PatientID': '123456',
                'StudyInstanceUID': '1.2.840.113619.2.55.3.604688.98765432.9876543210',
            },
            path.name,
        ).resolve()

        # try again after explicitly touching the expected path
        format_str = '/tmp/path/%(PatientID)s/%(StudyInstanceUID)s'
        expected_path = Path(
            format_str
            % {
                'PatientID': '123456',
                'StudyInstanceUID': '1.2.840.113619.2.55.3.604688.98765432.9876543210',
            },
            path.name,
        )
        expected_path.parent.mkdir(parents=True, exist_ok=True)
        expected_path.touch()
        with pytest.raises(FileExistsError):
            result = resolve_path(path, keys, format_str, truncate=0)

    def test_resolve_path_with_absolute_path(self, dicom_test_file) -> None:
        path = dicom_test_file
        keys = {'PatientID', 'StudyInstanceUID'}
        format_str = '/resolved/path/%(PatientID)s/%(StudyInstanceUID)s'
        result = resolve_path(path, keys, format_str, truncate=5)
        assert result[0] == path
        assert (
            result[1]
            == Path(
                format_str % {'PatientID': '123456', 'StudyInstanceUID': '43210'}, path.name
            ).resolve()
        )

    def test_resolve_path_with_nonexistent_path(self, dicom_test_file) -> None:
        path = Path('/nonexistent/path/to/file.dcm')
        keys = {'PatientID', 'StudyInstanceUID'}
        format_str = '/resolved/path/%(PatientID)s/%(StudyInstanceUID)s'
        with pytest.raises(FileNotFoundError):
            resolve_path(path, keys, format_str, truncate = 5)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict

from imgtools.coretypes import MedImage
from imgtools.io.readers import read_dicom_series
from imgtools.loggers import logger

if TYPE_CHECKING:
    import SimpleITK as sitk

__all__ = ["Scan"]


def read_dicom_scan(
    path: str,
    series_id: str | None = None,
    recursive: bool = False,
    file_names: list[str] | None = None,
    **kwargs: Any,  # noqa
) -> Scan:
    """Read a DICOM scan from a directory.

    Parameters
    ----------
    path : str
        Path to the directory containing the DICOM files.
    series_id : str | None, optional
        Series ID to read, by default None
    recursive : bool, optional
        Whether to read the files recursively, by default False
    file_names : list[str] | None, optional
        List of file names to read, by default None
    **kwargs : Any
        Unused keyword arguments.
    Returns
    -------
    Scan
        The read scan.
    """
    return Scan.from_dicom(
        path,
        series_id=series_id,
        recursive=recursive,
        file_names=file_names,
        **kwargs,
    )


class Scan(MedImage):
    metadata: Dict[str, Any]

    def __init__(self, image: sitk.Image, metadata: Dict[str, Any]) -> None:
        super().__init__(image)
        self.metadata = metadata
        self._fix_direction()

    def _fix_direction(self) -> None:
        """Validation of the image.
        sitk 2.4.0 now considers the `SpacingBetweenSlices` when determining
        the direction of the image. Sometimes, this is negative, which
        causes the z-axis to be flipped:
        - Direction([1.00,0.00,0.00], [0.00,1.00,0.00], [0.00,0.00,-1.00])

        So we need to check if the image is flipped and correct it.
        See https://github.com/SimpleITK/SimpleITK/issues/2214

        Note: if this is undesired behaviorm, we can have a config option to
        disable this manual fixing
        """

        slice_spacing = float(self.metadata.get("SpacingBetweenSlices") or 0)
        if (not slice_spacing) or (slice_spacing >= 0):
            return

        # just because metadata says that the spacing is negative, still
        # check if the image is flipped
        # in case  Scan was created from another (correct) Scan's transform
        if self.direction.to_matrix()[2][2] > 0:
            return

        warnmsg = (
            f"Scan has negative SpacingBetweenSlices: {slice_spacing}. "
            "Manually correcting the direction."
        )
        logger.debug(
            warnmsg,
            spacing=self.spacing,
            direction=self.direction,
        )

        self.SetDirection(self.direction.flip_axis(2))
        logger.debug(
            "Scan direction corrected.",
            spacing=self.spacing,
            direction=self.direction,
        )

    @classmethod
    def from_dicom(
        cls,
        path: str,
        series_id: str | None = None,
        recursive: bool = False,
        file_names: list[str] | None = None,
        **kwargs: Any,  # noqa
    ) -> Scan:
        """Read a DICOM scan from a directory.

        Parameters
        ----------
        path : str
            Path to the directory containing the DICOM files.
        series_id : str | None, optional
            Series ID to read, by default None
        recursive : bool, optional
            Whether to read the files recursively, by default False
        file_names : list[str] | None, optional
            List of file names to read, by default None
        **kwargs : Any
            Unused keyword arguments.
        Returns
        -------
        Scan
            The read scan.
        """
        image, metadata = read_dicom_series(
            path,
            series_id=series_id,
            recursive=recursive,
            file_names=file_names,
            **kwargs,
        )

        return cls(image, metadata)

    def __repr__(self) -> str:  # type: ignore
        # convert metadata and img_stats to string
        # with angulated brackets
        retr_str = "Scan<"
        for k, v in self.metadata.items():
            retr_str += f"\n\t{k}={v}" if v else ""
        retr_str += "\n>"
        return retr_str

    def __rich_repr__(self):  # type: ignore[no-untyped-def] # noqa: ANN204
        yield "modality", self.metadata.get("Modality", "Unknown")
        yield from super().__rich_repr__()


if __name__ == "__main__":  # pragma: no cover
    from rich import print  # noqa: A004

    directory = "data/4D-Lung/113_HM10395/CT_Series-00173972"
    scan = read_dicom_scan(directory)
    print(f"{scan!r}")  # noqa
    print(scan.fingerprint)



from dataclasses import dataclass, field
from pathlib import Path
from typing import ClassVar

import numpy as np
import SimpleITK as sitk

from imgtools.coretypes import MedImage
from imgtools.loggers import logger
from imgtools.utils import truncate_uid

from .abstract_base_writer import AbstractBaseWriter, ExistingFileMode


class NiftiWriterError(Exception):
    """Base exception for NiftiWriter errors."""

    pass


class NiftiWriterValidationError(NiftiWriterError):
    """Raised when validation of writer configuration fails."""

    pass


class NiftiWriterIOError(NiftiWriterError):
    """Raised when I/O operations fail."""

    pass


@dataclass
class NIFTIWriter(AbstractBaseWriter[sitk.Image | np.ndarray]):
    """Class for managing file writing with customizable paths and filenames for NIFTI files.

    This class extends the AbstractBaseWriter to provide specialized functionality
    for writing NIFTI image files. It supports both SimpleITK Image objects and numpy arrays
    as input data types.

    Attributes
    ----------
    compression_level : int, default=9
        Compression level (0-9). Higher means better compression but slower writing.
        Value must be between MIN_COMPRESSION_LEVEL (0) and MAX_COMPRESSION_LEVEL (9).
    truncate_uids_in_filename : int, default=8
        Many DICOM files have long UIDs in the filename. If used in the filename format,
        this will truncate the UID to the **last** `truncate_uids_in_filename`
        characters.
        A value of 0 means **no truncation**.
    VALID_EXTENSIONS : ClassVar[list[str]]
        List of valid file extensions for NIFTI files (".nii", ".nii.gz").
    MAX_COMPRESSION_LEVEL : ClassVar[int]
        Maximum allowed compression level (9).
    MIN_COMPRESSION_LEVEL : ClassVar[int]
        Minimum allowed compression level (0).

    Inherited Attributes
    -------------------
    root_directory : Path
        Root directory where files will be saved. This directory will be created
        if it doesn't exist and `create_dirs` is True.
    filename_format : str
        Format string defining the directory and filename structure.
        Supports placeholders for context variables enclosed in curly braces.
        Example: '{subject_id}_{date}/{disease}.nii.gz'
    create_dirs : bool, default=True
        Creates necessary directories if they don't exist.
    existing_file_mode : ExistingFileMode, default=ExistingFileMode.FAIL
        Behavior when a file already exists.
        Options: OVERWRITE, SKIP, FAIL
    sanitize_filenames : bool, default=True
        Replaces illegal characters from filenames with underscores.
    context : Dict[str, Any], default={}
        Internal context storage for pre-checking.
    index_filename : Optional[str], default=None
        Name of the index file to track saved files.
        If an absolute path is provided, it will be used as is.
        If not provided, it will be saved in the root directory with the format
        of {root_directory.name}_index.csv.
    overwrite_index : bool, default=False
        Overwrites the index file if it already exists.
    absolute_paths_in_index : bool, default=False
        If True, saves absolute paths in the index file.
        If False, saves paths relative to the root directory.
    pattern_resolver : PatternResolver
        Instance used to handle filename formatting with placeholders.

    Notes
    -----
    When using this class, ensure your filename_format ends with one of the VALID_EXTENSIONS.
    The class validates the compression level and filename format during initialization.
    """

    compression_level: int = field(default=9)
    truncate_uids_in_filename: int = field(default=8)

    # Make extensions immutable
    VALID_EXTENSIONS: ClassVar[list[str]] = [
        ".nii",
        ".nii.gz",
    ]
    MAX_COMPRESSION_LEVEL: ClassVar[int] = 9
    MIN_COMPRESSION_LEVEL: ClassVar[int] = 0

    def __post_init__(self) -> None:
        """Validate writer.

        Parent class will create the output directory, validate filename pattern, and update
        ExtingFileMode if necessary.
        """
        super().__post_init__()

        if (
            not self.MIN_COMPRESSION_LEVEL
            <= self.compression_level
            <= self.MAX_COMPRESSION_LEVEL
        ):
            msg = (
                f"Invalid compression level {self.compression_level}. "
                f"Must be between {self.MIN_COMPRESSION_LEVEL} and "
                f"{self.MAX_COMPRESSION_LEVEL}."
            )
            raise NiftiWriterValidationError(msg)

        if not any(
            self.filename_format.endswith(ext) for ext in self.VALID_EXTENSIONS
        ):
            msg = (
                f"Invalid filename format {self.filename_format}. "
                f"Must end with one of {self.VALID_EXTENSIONS}."
            )
            raise NiftiWriterValidationError(msg)

    def save(self, data: sitk.Image | np.ndarray, **kwargs: object) -> Path:
        """Write the SimpleITK image to a NIFTI file.

        Parameters
        ----------
        data : sitk.Image | np.ndarray
            The SimpleITK image or numpy array to save
        **kwargs : object
            Additional formatting parameters for the output path

        Returns
        -------
        Path
            Path to the saved file

        Raises
        ------
        NiftiWriterIOError
            If writing fails
        NiftiWriterValidationError
            If the input data is invalid
        """
        match data:
            case sitk.Image():
                image = data
            case np.ndarray():
                image = sitk.GetImageFromArray(data)
            case _:
                msg = "Input must be a SimpleITK Image or a numpy array"
                raise NiftiWriterValidationError(msg)

        # if the object has the 'fingerprint' property, update the context
        if hasattr(data, "serialized_fingerprint"):
            self.set_context(**data.serialized_fingerprint)
        elif isinstance(data, MedImage):
            # if there is no fingerprint, this is unexpected
            # this is an issue now since we auto keep context between
            # saves, so this could lead to using the fingerprint
            # of the previous save
            logger.error(
                "No fingerprint found in the writer object. "
                "This is unexpected and may indicate a bug. "
                "The fingerprint fields in the index might be incorrect."
            )
            # TODO:: think of a better way to handle this
            self.clear_context()

        # TODO:: think of a better way to handle the truncate_uids_in_filename
        if self.truncate_uids_in_filename:
            truncated_kwargs = {
                k: truncate_uid(str(v), self.truncate_uids_in_filename)
                if k.lower().endswith("uid")
                else v
                for k, v in kwargs.items()
            }
            out_path = self.resolve_path(**truncated_kwargs)
            # need to update the context with the old kwargs
            # because it will be used in the index, and we dont want
            # to truncate the UIDs in the index
            self.set_context(**kwargs)
        else:
            out_path = self.resolve_path(**kwargs)

        if (
            out_path.exists()  # check if it exists
            # This will only be true if SKIP,
            # OVERWRITE would have deleted the file
            and self.existing_file_mode == ExistingFileMode.SKIP
        ):
            logger.debug("File exists, skipping.", out_path=out_path)
            return out_path

        try:
            sitk.WriteImage(
                image,
                out_path.as_posix(),
                useCompression=True,
                compressionLevel=self.compression_level,
            )
        except Exception as e:
            msg = f"Error writing image to file {out_path}: {e}"
            raise NiftiWriterIOError(msg) from e

        self.add_to_index(
            out_path,
            filepath_column="filepath",
            replace_existing=out_path.exists(),
        )

        return out_path
